from typing import Type, Union, Optional, TypeVar, Collection

from redis import Redis
from redis.client import Pipeline

from assimilator.redis_.database import RedisModel
from assimilator.core.patterns.error_wrapper import ErrorWrapper
from assimilator.core.database import (
    SpecificationList,
    SpecificationType,
    Repository,
    LazyCommand,
)
from assimilator.internal.database.specifications import InternalSpecificationList
from assimilator.internal.database.models_utils import dict_to_models
from assimilator.core.database.exceptions import (
    DataLayerError,
    NotFoundError,
    InvalidQueryError,
    MultipleResultsError,
)

RedisModelT = TypeVar("RedisModelT", bound=RedisModel)


class RedisRepository(Repository):
    session: Redis
    transaction: Union[Pipeline, Redis]
    model: Type[RedisModelT]

    def __init__(
        self,
        session: Redis,
        model: Type[RedisModelT],
        initial_query: Optional[str] = '',
        specifications: Type[SpecificationList] = InternalSpecificationList,
        error_wrapper: Optional[ErrorWrapper] = None,
        use_double_filter: bool = True,
    ):
        super(RedisRepository, self).__init__(
            session=session,
            model=model,
            initial_query=initial_query,
            specifications=specifications,
            error_wrapper=error_wrapper or ErrorWrapper(
                default_error=DataLayerError,
                skipped_errors=(NotFoundError,)
            )
        )
        self.transaction = session
        self.use_double_specifications = use_double_filter

    def get(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Optional[str] = None,
    ) -> Union[LazyCommand[RedisModelT], RedisModelT]:
        query = self._apply_specifications(
            query=initial_query,
            specifications=specifications,
        ) or '*'

        found_objects = self.session.mget(self.session.keys(query))

        if not all(found_objects):
            raise NotFoundError()

        parsed_objects = list(self._apply_specifications(
            query=[self.model.loads(found_object) for found_object in found_objects],
            specifications=specifications,
        ))

        if not parsed_objects:
            raise NotFoundError()
        elif len(parsed_objects) != 1:
            raise MultipleResultsError()

        return parsed_objects[0]

    def filter(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Optional[str] = None,
    ) -> Union[LazyCommand[Collection[RedisModelT]], Collection[RedisModelT]]:
        if self.use_double_specifications and specifications:
            key_name = self._apply_specifications(
                query=initial_query,
                specifications=specifications,
            ) or "*"
        else:
            key_name = "*"

        models = self.session.mget(self.session.keys(key_name))
        return list(self._apply_specifications(specifications=specifications, query=[
            self.model.loads(value) for value in models
        ]))

    def save(self, obj: Optional[RedisModelT] = None, **obj_data) -> RedisModelT:
        if obj is None:
            obj = self.model(**dict_to_models(data=obj_data, model=self.model))

        self.transaction.set(
            name=obj.id,
            value=obj.json(),
            ex=getattr(obj, 'expire_in', None),     # for Pydantic model compatability
            px=getattr(obj, 'expire_in_px', None),
            nx=getattr(obj, 'only_create', False),
            xx=getattr(obj, 'only_update', False),
            keepttl=getattr(obj, 'keep_ttl', False),
        )
        return obj

    def delete(self, obj: Optional[RedisModelT] = None, *specifications: SpecificationType) -> None:
        obj, specifications = self._check_obj_is_specification(obj, specifications)

        if specifications:
            self.transaction.delete(*[str(model.id) for model in self.filter(*specifications)])
        elif obj is not None:
            self.transaction.delete(obj.id)

    def update(
        self,
        obj: Optional[RedisModelT] = None,
        *specifications: SpecificationType,
        **update_values,
    ) -> None:
        obj, specifications = self._check_obj_is_specification(obj, specifications)

        if specifications:
            if not update_values:
                raise InvalidQueryError(
                    "You did not provide any update_values "
                    "to the update() yet provided specifications"
                )

            models = self.filter(*specifications, lazy=False)
            updated_models = {}

            for model in models:
                model.__dict__.update(update_values)
                updated_models[str(model.id)] = model.json()

            self.transaction.mset(updated_models)

        elif obj is not None:
            obj.only_update = True
            self.save(obj)

    def is_modified(self, obj: RedisModelT) -> None:
        return self.get(self.specifications.filter(obj.id), lazy=False) == obj

    def refresh(self, obj: RedisModelT) -> None:
        fresh_obj = self.get(self.specifications.filter(obj.id), lazy=False)

        for key, value in fresh_obj.dict().items():
            setattr(obj, key, value)

    def count(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Optional[str] = None,
    ) -> Union[LazyCommand[int], int]:
        if not specifications:
            return self.session.dbsize()

        filter_query = self._apply_specifications(
            query=initial_query,
            specifications=specifications,
        )
        return len(self.session.keys(filter_query))


__all__ = [
    'RedisRepository',
]



from typing import Type, Union, Optional, TypeVar, List

from assimilator.core.patterns.error_wrapper import ErrorWrapper
from assimilator.internal.database.error_wrapper import InternalErrorWrapper
from assimilator.core.database import (
    Repository,
    SpecificationType,
    LazyCommand,
    InvalidQueryError,
    BaseModel,
    NotFoundError,
)
from assimilator.internal.database.specifications import InternalSpecificationList
from assimilator.core.database import MultipleResultsError
from internal.database.models_utils import dict_to_models

ModelT = TypeVar("ModelT", bound=BaseModel)


class InternalRepository(Repository):
    session: dict
    model: Type[ModelT]

    def __init__(
        self,
        session: dict,
        model: Type[ModelT],
        initial_query: Optional[str] = '',
        specifications: Type[InternalSpecificationList] = InternalSpecificationList,
        error_wrapper: Optional[ErrorWrapper] = None,
    ):
        super(InternalRepository, self).__init__(
            model=model,
            session=session,
            initial_query=initial_query,
            specifications=specifications,
            error_wrapper=error_wrapper or InternalErrorWrapper(),
        )

    def get(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Optional[str] = None,
    ) -> Union[LazyCommand[ModelT], ModelT]:
        query = self._apply_specifications(
            query=initial_query,
            specifications=specifications,
        )

        if query:   # Dict key was not provided, we must use other search parameters
            return self.session[query]

        found_models = list(self._apply_specifications(
            query=self.session.values(),
            specifications=specifications,
        ))

        if not found_models:
            raise NotFoundError()
        elif len(found_models) != 1:
            raise MultipleResultsError()

        return found_models[0]

    def filter(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Optional[str] = None,
    ) -> Union[LazyCommand[List[ModelT]], List[ModelT]]:
        return list(self._apply_specifications(
            query=self.session.values(),
            specifications=specifications,
        ))

    def save(self, obj: Optional[ModelT] = None, **obj_data) -> ModelT:
        if obj is None:
            obj = self.model(**dict_to_models(data=obj_data, model=self.model))

        self.session[obj.id] = obj
        return obj

    def delete(self, obj: Optional[ModelT] = None, *specifications: SpecificationType) -> None:
        obj, specifications = self._check_obj_is_specification(obj, specifications)

        if specifications:
            for model in self.filter(*specifications, lazy=True):
                del self.session[model.id]
        elif obj is not None:
            del self.session[obj.id]

    def update(
        self,
        obj: Optional[ModelT] = None,
        *specifications: SpecificationType,
        **update_values,
    ) -> None:
        obj, specifications = self._check_obj_is_specification(obj, specifications)

        if specifications:
            if not update_values:
                raise InvalidQueryError(
                    "You did not provide any update_values "
                    "to the update() yet provided specifications"
                )

            for model in self.filter(*specifications, lazy=True):
                model.__dict__.update(update_values)
                self.save(model)

        elif obj is not None:
            self.save(obj)

    def is_modified(self, obj: ModelT) -> bool:
        return self.get(self.specs.filter(id=obj.id)) == obj

    def refresh(self, obj: ModelT) -> None:
        fresh_obj = self.get(self.specs.filter(id=obj.id), lazy=False)
        obj.__dict__.update(fresh_obj.__dict__)

    def count(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Optional[str] = None,
    ) -> Union[LazyCommand[int], int]:
        if specifications:
            return len(list(self._apply_specifications(  # We do not call filter() for list() optimization
                query=self.session.values(),
                specifications=specifications,
            )))
        return len(self.session)


__all__ = [
    'InternalRepository',
]



from assimilator.alchemy.database.repository import AlchemyRepository
from assimilator.alchemy.database.error_wrapper import AlchemyErrorWrapper
from assimilator.core.database.unit_of_work import UnitOfWork
from assimilator.core.patterns.error_wrapper import ErrorWrapper


class AlchemyUnitOfWork(UnitOfWork):
    repository: AlchemyRepository

    def __init__(self, repository: AlchemyRepository, error_wrapper: ErrorWrapper = None):
        super(AlchemyUnitOfWork, self).__init__(
            repository=repository,
            error_wrapper=error_wrapper or AlchemyErrorWrapper(),
        )

    def begin(self):
        self.repository.session.begin()

    def rollback(self):
        self.repository.session.rollback()

    def close(self):
        pass

    def commit(self):
        self.repository.session.commit()


__all__ = [
    'AlchemyUnitOfWork',
]



from functools import wraps
from typing import Dict, Type, Optional, Callable, Container


class ErrorWrapper:
    def __init__(
        self,
        error_mappings: Optional[Dict[Type[Exception], Type[Exception]]] = None,
        default_error: Optional[Type[Exception]] = None,
        skipped_errors: Optional[Container[Type[Exception]]] = None,
    ):
        self.error_mappings = error_mappings or {}
        self.default_error = default_error
        self.skipped_errors = {
            *(skipped_errors or set()),
            KeyboardInterrupt,
            SystemExit,
            *self.error_mappings.values(),  # we want to skip all the mapped values as they are already fixed
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Type[Exception], exc_val: Exception, exc_tb):
        if exc_val is None:
            return True
        elif exc_type in self.skipped_errors:
            return False

        wrapped_error = self.error_mappings.get(exc_type)

        if wrapped_error is not None:
            raise wrapped_error(exc_val) from exc_val
        elif self.default_error is not None:
            raise self.default_error(exc_val) from exc_val

        return False   # No wrapping error was found

    def decorate(self, func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        wrapper: func
        return wrapper

    def __str__(self):
        return f"{type(self).__name__}({self.error_mappings})"


__all__ = ['ErrorWrapper']



from itertools import zip_longest
from typing import Collection, Optional, Iterable, Any

from sqlalchemy.orm import load_only, joinedload
from sqlalchemy import column, desc, and_, or_, not_, Select

from assimilator.alchemy.database.model_utils import get_model_from_relationship
from assimilator.alchemy.database.specifications.filtering_options import AlchemyFilteringOptions
from assimilator.core.database.specifications import (
    specification,
    SpecificationList,
    SpecificationType,
    FilterSpecification,
)


class AlchemyFilter(FilterSpecification):
    filtering_options_cls = AlchemyFilteringOptions

    def __or__(self, other: 'AlchemyFilter') -> SpecificationType:
        return AlchemyFilter(or_(*self.filters, *other.filters))

    def __and__(self, other: 'AlchemyFilter') -> SpecificationType:
        return AlchemyFilter(and_(*self.filters, *other.filters))

    def __invert__(self):
        return AlchemyFilter(not_(*self.filters))

    def apply(self, query: Select, **context: Any) -> Select:
        return query.filter(*self.filters)


alchemy_filter = AlchemyFilter


@specification
def alchemy_order(*clauses: str, query: Select, **_) -> Select:
    parsed_clauses = []

    for clause in clauses:
        if clause.startswith("-"):
            parsed_clauses.append(desc(column(clause[1:])))
        else:
            parsed_clauses.append(clause)

    return query.order_by(*parsed_clauses)


@specification
def alchemy_paginate(
    *,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    query: Select,
    **_,
) -> Select:
    if offset is not None:
        query = query.offset(offset)
    if limit is not None:
        query = query.limit(limit)

    return query


@specification
def alchemy_join(
    *targets: Collection,
    join_args: Iterable[dict] = None,
    query: Select,
    model,
    **_,
) -> Select:
    for target, join_data in zip_longest(targets, (join_args or {}), fillvalue=dict()):
        if not target:
            continue

        if isinstance(target, str):
            entities = target.split(".")
            target = model

            for entity in entities:
                target, _ = get_model_from_relationship(
                    model=target,
                    relationship_name=entity,
                )

        query = query.join(target, **join_data).add_columns(target).select_from(model)

    return query


@specification
def alchemy_only(
    *only_fields: Iterable[str],
    query: Select,
    **_,
):
    return query.options(load_only(*only_fields))


class AlchemySpecificationList(SpecificationList):
    filter = AlchemyFilter
    order = alchemy_order
    paginate = alchemy_paginate
    join = alchemy_join
    only = alchemy_only


__all__ = [
    'AlchemySpecificationList',
    'alchemy_filter',
    'AlchemyFilter',
    'alchemy_order',
    'alchemy_paginate',
    'alchemy_join',
    'alchemy_only',
]



from typing import Callable, Any, Optional, Collection

from assimilator.mongo.database.specifications import MongoFilteringOptions
from assimilator.core.database import SpecificationList, FilterSpecification, specification


class MongoFilter(FilterSpecification):
    filters: dict
    filtering_options_cls = MongoFilteringOptions

    def __init__(self, *filters, **named_filters):
        super(MongoFilter, self).__init__(*filters, **named_filters)
        parsed_filters = {}

        for filter_ in self.filters:
            parsed_filters.update(filter_)

        self.filters = parsed_filters

    def get_parsed_filter(self, filter_func: Callable, field: str, value: Any) -> dict:
        return filter_func(field, value)

    def __or__(self, other: 'FilterSpecification') -> 'FilterSpecification':
        return MongoFilter({"$or": [self.filters, other.filters]})

    def __and__(self, other: 'FilterSpecification') -> 'FilterSpecification':
        return MongoFilter({**self.filters, **other.filters})

    def __invert__(self) -> 'MongoFilter':
        inverted_filters = []

        for column, value in self.filters.items():
            inverted_filters.append({column: {"$not": value}})

        return MongoFilter(*inverted_filters)

    def apply(self, query: dict, **context: Any) -> dict:
        query['filter'] = {**query.get('filter', {}), **self.filters}
        return query


mongo_filter = MongoFilter


@specification
def mongo_order(*clauses: str, query: dict, **_) -> dict:
    query['sort'] = {
        **query.get('sort', {}),
        **{column: -1 if column.startswith("-") else 1 for column in clauses}
    }
    return query


@specification
def mongo_paginate(
    *,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    query: dict,
    **_,
) -> dict:
    if offset is not None:
        query['skip'] = offset
    if limit is not None:
        query['limit'] = limit

    return query


@specification
def mongo_join(*targets: Collection, query: dict, **join_args: dict) -> dict:
    return query


@specification
def mongo_only(*only_fields: str, query: dict, **_) -> dict:
    query['projection'] = only_fields
    return query


class MongoSpecificationList(SpecificationList):
    filter = MongoFilter
    order = mongo_order
    paginate = mongo_paginate
    join = mongo_join
    only = mongo_only


__all__ = [
    'MongoSpecificationList',
    'MongoFilter',
    'mongo_filter',
    'mongo_order',
    'mongo_paginate',
    'mongo_join',
    'mongo_only',
]



from assimilator.alchemy.database.repository import *
from assimilator.alchemy.database.specifications.specifications import *
from assimilator.alchemy.database.specifications.filtering_options import *
from assimilator.alchemy.database.unit_of_work import *



from abc import ABC, abstractmethod
from typing import Optional

from assimilator.core.database.repository import Repository
from assimilator.core.patterns import ErrorWrapper


class UnitOfWork(ABC):
    error_wrapper: ErrorWrapper = ErrorWrapper()

    def __init__(self, repository: Repository, error_wrapper: Optional[ErrorWrapper] = None):
        self.repository = repository
        if error_wrapper is not None:
            self.error_wrapper = error_wrapper

        self.begin = self.error_wrapper.decorate(self.begin)
        self.rollback = self.error_wrapper.decorate(self.rollback)
        self.commit = self.error_wrapper.decorate(self.commit)
        self.close = self.error_wrapper.decorate(self.close)

    @abstractmethod
    def begin(self):
        raise NotImplementedError()

    @abstractmethod
    def rollback(self):
        raise NotImplementedError()

    @abstractmethod
    def commit(self):
        raise NotImplementedError()

    @abstractmethod
    def close(self):
        raise NotImplementedError()

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self.rollback()
            self.close()
            raise exc_val
        else:
            self.close()

    def __str__(self):
        return f"{self.__class__.__name__}({self.repository.model})"

    def __repr__(self):
        return str(self)


__all__ = [
    'UnitOfWork',
]



from typing import Protocol, TypeVar, Optional, Iterable, Any

QueryT = TypeVar("QueryT")


class OrderSpecificationProtocol(Protocol):
    def __call__(self, *clauses: str) -> QueryT:
        ...


class PaginateSpecificationProtocol(Protocol):
    def __call__(self, *, limit: Optional[int] = None, offset: Optional[int] = None):
        ...


class JoinSpecificationProtocol(Protocol):
    def __call__(self, *targets: Any, join_args: Iterable[dict] = None):
        ...


class OnlySpecificationProtocol(Protocol):
    def __call__(self, *only_fields: Iterable[str]) -> Iterable[QueryT]:
        ...


__all__ = [
    'OrderSpecificationProtocol',
    'PaginateSpecificationProtocol',
    'JoinSpecificationProtocol',
    'OnlySpecificationProtocol',
]



from typing import Type, Union, Optional, TypeVar, Collection

from sqlalchemy import func, select, update, delete, Delete
from sqlalchemy.orm import Session, Query
from sqlalchemy.inspection import inspect

from assimilator.alchemy.database.model_utils import dict_to_models
from assimilator.core.patterns.error_wrapper import ErrorWrapper
from assimilator.core.database.exceptions import InvalidQueryError
from assimilator.alchemy.database.error_wrapper import AlchemyErrorWrapper
from assimilator.alchemy.database.specifications import AlchemySpecificationList
from assimilator.core.database import Repository, LazyCommand, SpecificationType


AlchemyModelT = TypeVar("AlchemyModelT")


class AlchemyRepository(Repository):
    session: Session
    model: Type[AlchemyModelT]

    def __init__(
        self,
        session: Session,
        model: Type[AlchemyModelT],
        initial_query: Query = None,
        specifications: Type[AlchemySpecificationList] = AlchemySpecificationList,
        error_wrapper: Optional[ErrorWrapper] = None,
    ):
        super(AlchemyRepository, self).__init__(
            session=session,
            model=model,
            initial_query=initial_query if initial_query is not None else select(model),
            specifications=specifications,
            error_wrapper=error_wrapper or AlchemyErrorWrapper(),
        )

    def get(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Query = None,
    ) -> Union[AlchemyModelT, LazyCommand[AlchemyModelT]]:
        query = self._apply_specifications(
            query=initial_query,
            specifications=specifications,
        )
        return self.session.execute(query).one()[0]

    def filter(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Query = None,
    ) -> Union[Collection[AlchemyModelT], LazyCommand[Collection[AlchemyModelT]]]:
        query = self._apply_specifications(
            query=initial_query,
            specifications=specifications,
        )
        return [result[0] for result in self.session.execute(query)]

    def update(
        self,
        obj: Optional[AlchemyModelT] = None,
        *specifications: SpecificationType,
        **update_values,
    ) -> None:
        obj, specifications = self._check_obj_is_specification(obj, specifications)

        if specifications:
            if not update_values:
                raise InvalidQueryError(
                    "You did not provide any update_values "
                    "to the update() yet provided specifications"
                )

            query = self._apply_specifications(
                query=update(self.model),
                specifications=specifications,
            )

            self.session.execute(
                query.values(update_values).execution_options(synchronize_session=False)
            )

        elif obj is not None:
            if obj not in self.session:
                obj = self.session.merge(obj)

            self.session.add(obj)

    def save(self, obj: Optional[AlchemyModelT] = None, **data) -> AlchemyModelT:
        if obj is None:
            obj = self.model(**dict_to_models(data=data, model=self.model))

        self.session.add(obj)
        return obj

    def refresh(self, obj: AlchemyModelT) -> None:
        if obj not in self.session:
            obj = self.session.merge(obj)

        self.session.refresh(obj)

    def delete(self, obj: Optional[AlchemyModelT] = None, *specifications: SpecificationType) -> None:
        obj, specifications = self._check_obj_is_specification(obj, specifications)

        if specifications:
            query: Delete = self._apply_specifications(
                query=delete(self.model),
                specifications=specifications,
            )
            self.session.execute(query.execution_options(synchronize_session=False))
        elif obj is not None:
            self.session.delete(obj)

    def is_modified(self, obj: AlchemyModelT) -> bool:
        return obj in self.session and self.session.is_modified(obj)

    def count(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Query = None
    ) -> Union[LazyCommand[int], int]:
        primary_keys = inspect(self.model).primary_key

        if not primary_keys:
            raise InvalidQueryError(
                "Your repository model does not have any primary keys. We cannot use count()"
            )

        return self.get(
            *specifications,
            lazy=False,
            initial_query=initial_query or select(func.count(getattr(self.model, primary_keys[0].name))),
        )


__all__ = [
    'AlchemyRepository',
]



from typing import ClassVar

from bson import ObjectId

from assimilator.core.database.models import BaseModel


class MongoModel(BaseModel):
    upsert: bool = False
    id: ObjectId

    class AssimilatorConfig(BaseModel.AssimilatorConfig):
        collection: ClassVar[str]
        autogenerate_id: ClassVar[bool] = True

    class Config:
        exclude = {
            'collection': True,
            'upsert': True,
        }
        arbitrary_types_allowed = True

    def __init__(self, **kwargs):
        super(MongoModel, self).__init__(**kwargs)

    def generate_id(self, **kwargs) -> ObjectId:
        return ObjectId()


__all__ = ['MongoModel']



from typing import Optional

from assimilator.core.database.models import BaseModel


class RedisModel(BaseModel):
    expire_in: Optional[int] = None
    expire_in_px: Optional[int] = None
    only_update: Optional[bool] = False   # Same as xx in redis set. Only set if key exists
    only_create: Optional[bool] = False   # Same as nx in redis set. Only set if key does not exist
    keep_ttl: Optional[bool] = False

    class Config:
        exclude = {
            'expire_in': True,
            'expire_in_px': True,
            'only_update': True,
            'only_create': True,
            'keep_ttl': True,
        }


__all__ = [
    'RedisModel',
]



from functools import wraps
from typing import Union, Callable, Iterable, TypeVar, Generic, Iterator

T = TypeVar("T")


class LazyCommand(Generic[T]):
    def __init__(self, command: Callable, *args, **kwargs):
        self.command = command
        self.args = args
        self.kwargs = kwargs
        self._results: T = None

    def __call__(self) -> Union[T]:
        if self._results is not None:
            return self._results

        self._results = self.command(*self.args, **self.kwargs)
        return self._results

    def __iter__(self) -> Iterator[T]:
        results = self()

        if not isinstance(results, Iterable):  # get() command
            raise StopIteration("Results are not iterable")

        return iter(results)  # filter() command

    def __eq__(self, other):
        return self() == other

    def __gt__(self, other):
        return self() > other

    def __getattr__(self, item):
        result = self()
        return getattr(result, item)

    def __bool__(self):
        return bool(self())

    def __str__(self):
        return f"Lazy<{self.command}(*{self.args}, **{self.kwargs})>"

    def __repr__(self):
        return str(self)

    @staticmethod
    def decorate(func: Callable) -> Callable:

        @wraps(func)
        def lazy_wrapper(*args, lazy: bool = False, **kwargs) -> Union[LazyCommand[T], T]:
            if lazy:
                return LazyCommand(
                    func,
                    *args,
                    lazy=False,
                    **kwargs,
                )

            return func(*args, **kwargs)

        lazy_wrapper: func
        return lazy_wrapper


__all__ = ['LazyCommand']



from assimilator.core.database.exceptions import DataLayerError, NotFoundError
from assimilator.core.patterns.error_wrapper import ErrorWrapper


class InternalErrorWrapper(ErrorWrapper):
    def __init__(self):
        super(InternalErrorWrapper, self).__init__(error_mappings={
            KeyError: NotFoundError,
            TypeError: NotFoundError,
        }, default_error=DataLayerError)


__all__ = ['InternalErrorWrapper']



from typing import Any, Callable

from assimilator.core.database import BaseModel
from assimilator.core.database.specifications import FilteringOptions
from assimilator.internal.database.specifications.internal_operator import (
    find_attribute, eq, gte, gt, lte, lt, is_, not_, like, regex,
)


class InternalFilteringOptions(FilteringOptions):
    def __init__(self, attr_finder: Callable[[Callable, str, Any], Callable[[BaseModel], bool]] = find_attribute):
        super(InternalFilteringOptions, self).__init__()
        self.attr_finder = attr_finder

    _eq = staticmethod(eq)
    _gt = staticmethod(gt)
    _gte = staticmethod(gte)
    _lt = staticmethod(lt)
    _lte = staticmethod(lte)
    _not = staticmethod(not_)
    _is = staticmethod(is_)
    _like = staticmethod(like)
    _regex = staticmethod(regex)


__all__ = [
    'InternalFilteringOptions',
    'find_attribute',
    "eq",
    "gte",
    "gt",
    "lte",
    "lt",
    "is_",
    "not_",
    "like",
    "regex",
]





class DataLayerError(Exception):
    """ Any error related to Repository, UnitOfWork, Model """


class NotFoundError(DataLayerError):
    """ Results are not found """


class InvalidQueryError(DataLayerError):
    """ The query to the data storage supplied was invalid """


class MultipleResultsError(InvalidQueryError):
    """ Repository get() function returned more than one result """



from typing import Callable, Union, List, Generator

from assimilator.core.database.models import BaseModel
from assimilator.core.database.specifications import FilterSpecification
from assimilator.internal.database.specifications.filtering_options import InternalFilteringOptions


QueryT = Union[str, List[BaseModel]]


class InternalFilter(FilterSpecification):
    filtering_options_cls = InternalFilteringOptions

    def __init__(self, *filters, **named_filters):
        self.text_filters = [filter_ for filter_ in filters if isinstance(filter_, str)]

        if named_filters.get('id'):
            self.text_filters.append(named_filters.pop('id'))

        super(InternalFilter, self).__init__(
            *(set(filters) - set(self.text_filters)),
            **named_filters,
        )

    def apply(self, query: QueryT, **context) -> Union[str, Generator[bool, bool, None]]:
        if isinstance(query, str):
            return f'{query}{"".join(self.text_filters)}'
        elif not self.filters:
            return query

        return (model for model in query if all(filter_func(model) for filter_func in self.filters))

    def __or__(self, other: 'InternalFilter') -> 'InternalFilter':
        return OrInternalFilter(first_spec=self, second_spec=other)

    def __and__(self, other: 'InternalFilter') -> 'InternalFilter':
        specification = InternalFilter(*self.filters, *other.filters)
        specification.text_filters = self.text_filters + other.text_filters
        return specification

    @staticmethod
    def __inversion_func(func: Callable):
        def __inversion_func_wrapper(field, value):
            return not func(field, value)
        return __inversion_func_wrapper

    def __invert__(self):
        return InternalFilter(*map(
            lambda func, field, value: (self.__inversion_func(func), field, value),
            self.filters,
        ))


class OrInternalFilter(InternalFilter):
    def __init__(self, first_spec: InternalFilter, second_spec: InternalFilter):
        super(OrInternalFilter, self).__init__()
        self.first_spec = first_spec
        self.second_spec = second_spec

    def apply(self, query: QueryT, **context) -> Union[str, QueryT]:
        if isinstance(query, str):
            return f'{query}{"".join(self.first_spec.text_filters + self.second_spec.text_filters)}'

        first_filter = self.first_spec.apply(query)
        second_filter = self.second_spec.apply(query)
        return list(set(first_filter) | set(second_filter))


__all__ = ['InternalFilter']



from typing import Iterable, Optional

from redis import Redis
from redis.client import PubSub

from assimilator.core.events import Event, ExternalEvent
from assimilator.core.events.events_bus import EventConsumer, EventProducer


class RedisEventConsumer(EventConsumer):
    def __init__(self, channels: Iterable[str], session: Redis):
        super(RedisEventConsumer, self).__init__()
        self.session = session
        self.channels = channels
        self._event_channel: Optional[PubSub] = None

    def close(self):
        self._event_channel.close()
        self._event_channel = None

    def start(self):
        self._event_channel = self.session.pubsub()
        self._event_channel.subscribe(*self.channels)

    def consume(self) -> Iterable[ExternalEvent]:
        message = self._event_channel.get_message(ignore_subscribe_messages=True)

        while message is not None:
            if message['type'] == 'message':
                yield ExternalEvent.loads(message['data'])

            message = self._event_channel.get_message(ignore_subscribe_messages=True)


class RedisEventProducer(EventProducer):
    def __init__(self, channel: str, session: Redis):
        self.session = session
        self.channel = channel

    def produce(self, event: Event):
        self.session.publish(self.channel, event.json())

    def start(self):
        pass

    def close(self):
        pass


__all__ = [
    'RedisEventConsumer',
    'RedisEventProducer',
]



from abc import ABC
from typing import Iterable

from assimilator.core.database.unit_of_work import UnitOfWork
from assimilator.core.events.events import Event
from assimilator.core.events.events_bus import EventProducer


class OutboxRelay(ABC):
    def __init__(self, uow: UnitOfWork, producer: EventProducer):
        self.uow = uow
        self.producer = producer

    def start(self):
        raise NotImplementedError("start() is not implemented")

    def acknowledge(self, events: Iterable[Event]):
        raise NotImplementedError("acknowledge() is not implemented")


__all__ = [
    'OutboxRelay',
]



from datetime import datetime
from typing import Any

from pydantic import Field

from assimilator.core.database.models import BaseModel


class Event(BaseModel):
    event_name: str = Field(allow_mutation=False)
    event_date: datetime = Field(default_factory=datetime.now)

    class Config:
        validate_assignment = True


class ExternalEvent(Event):
    """ The event type is unknown, so all the fields are in data """
    data: Any


class AckEvent(Event):
    acknowledged: bool = False


__all__ = [
    'Event',
    'ExternalEvent',
    'AckEvent',
]



from typing import Type, Optional

from sqlalchemy import Table
from sqlalchemy.orm import Query

from assimilator.alchemy.database.specifications import AlchemySpecificationList
from assimilator.alchemy.database.repository import AlchemyRepository
from assimilator.core.database import SpecificationList
from assimilator.core.patterns.error_wrapper import ErrorWrapper


class AlchemyOutboxRepository(AlchemyRepository):
    def __init__(
        self,
        session,
        event_model: Type[Table],
        model: Type[Table],
        initial_query: Optional[Query] = None,
        specifications: Type[SpecificationList] = AlchemySpecificationList,
        error_wrapper: ErrorWrapper = None,
    ):
        super(AlchemyOutboxRepository, self).__init__(
            session=session,
            initial_query=initial_query,
            model=model,
            specifications=specifications,
            error_wrapper=error_wrapper,
        )
        self.event_model = event_model

    def save(self, obj):
        super(AlchemyOutboxRepository, self).save(obj)
        super(AlchemyOutboxRepository, self).save(self.event_model(obj.outbox_event))


__all__ = [
    'AlchemyOutboxRepository',
]



from assimilator.core.events.events import Event
from assimilator.core.events.events_bus import EventConsumer, EventProducer


class InternalEventConsumer(EventConsumer):
    def __init__(self, event_storage: list):
        self.event_storage = event_storage

    def close(self):
        pass

    def start(self):
        pass

    def consume(self):
        while self.event_storage:
            yield self.event_storage.pop()


class InternalEventProducer(EventProducer):
    def __init__(self, event_storage: list):
        self.event_storage = event_storage

    def produce(self, event: Event):
        self.event_storage.append(event)

    def start(self):
        pass

    def close(self):
        pass


__all__ = [
    'InternalEventConsumer',
    'InternalEventProducer',
]



from typing import TypeVar, Type

from sqlalchemy import inspect


T = TypeVar("T")


def get_model_from_relationship(model: T, relationship_name: str):
    foreign_prop = getattr(model, relationship_name).property
    try:
        return foreign_prop.mapper_registry.class_, foreign_prop.uselist
    except AttributeError:
        return foreign_prop.entity.class_, foreign_prop.uselist


def dict_to_models(data: dict, model: Type[T]) -> T:
    for relationship in inspect(model).relationships.keys():
        foreign_data = data.get(relationship)
        if foreign_data is None:
            continue

        foreign_model, is_list = get_model_from_relationship(
            model=model,
            relationship_name=relationship,
        )

        if not is_list and isinstance(foreign_data, dict):
            foreign_data = dict_to_models(data=foreign_data, model=foreign_model)
            foreign_data = foreign_model(**foreign_data)
        elif is_list:
            foreign_models = (
                foreign_data for foreign_data in foreign_data
                if isinstance(foreign_data, dict)
            )

            for i, foreign_part in enumerate(foreign_models):
                foreign_part = dict_to_models(data=foreign_part, model=foreign_model)
                foreign_data[i] = foreign_model(**foreign_part)

        data[relationship] = foreign_data

    return data



from assimilator.core.database.repository import *
from assimilator.core.database.specifications.specifications import *
from assimilator.core.database.specifications.filtering_options import *
from assimilator.core.database.unit_of_work import *
from assimilator.core.database.exceptions import *
from assimilator.core.database.models import *



from bson.errors import BSONError
from pymongo.errors import DuplicateKeyError, InvalidOperation, WriteError

from assimilator.core.patterns import ErrorWrapper
from assimilator.core.exceptions import ParsingError
from assimilator.core.database import DataLayerError, InvalidQueryError, NotFoundError


class MongoErrorWrapper(ErrorWrapper):
    def __init__(self):
        super(MongoErrorWrapper, self).__init__(
            error_mappings={
                BSONError: ParsingError,
                DuplicateKeyError: InvalidQueryError,
                InvalidOperation: InvalidQueryError,
                WriteError: InvalidQueryError,
            },
            skipped_errors={NotFoundError},
            default_error=DataLayerError,
        )


__all__ = ['MongoErrorWrapper']



from typing import Type, Union

from pydantic import BaseModel as PydanticBaseModel

from assimilator.core.database.models import BaseModel


def get_model_relationship(model: Type[BaseModel], field_name: str) -> Union[Type[BaseModel], None]:
    try:
        return model.__fields__.get(field_name).type_
    except AttributeError:
        return None


def dict_to_models(data: dict, model: Type[BaseModel]) -> dict:
    for field_name, value in dict(data).items():
        field_type = get_model_relationship(model, field_name)
        if field_type is None:
            continue
        elif not issubclass(field_type, PydanticBaseModel):
            continue

        if not isinstance(value, dict):
            data[field_name] = [
                field_type(**dict_to_models(data=val_part, model=field_type))
                for val_part in value
            ]
        else:
            data[field_name] = field_type(**value)

    return data



from copy import deepcopy
from typing import Optional

from assimilator.core.database import UnitOfWork
from assimilator.core.database.repository import Repository
from assimilator.internal.database.error_wrapper import InternalErrorWrapper


class InternalUnitOfWork(UnitOfWork):
    def __init__(self, repository: Repository):
        super(InternalUnitOfWork, self).__init__(
            repository=repository,
            error_wrapper=InternalErrorWrapper(),
        )
        self._saved_data: Optional[dict] = None

    def begin(self):
        self._saved_data = self.repository.session
        self.repository.session = deepcopy(self._saved_data)

    def rollback(self):
        self.repository.session = self._saved_data

    def commit(self):
        self._saved_data.update(self.repository.session)

        for deleted_key in set(self._saved_data.keys()) - set(self.repository.session.keys()):
            del self._saved_data[deleted_key]

        self.repository.session = self._saved_data

    def close(self):
        self._saved_data = None


__all__ = [
    'InternalUnitOfWork',
]



import operator
import re
from functools import wraps
from numbers import Number
from typing import Any, Callable, Tuple, Set, List, Union, Literal, Dict

from assimilator.core.database import BaseModel, FILTERING_OPTIONS_SEPARATOR


Containers = (List, Set, Tuple, map)


def find_attribute(func: callable, field: str, value: Any) -> Callable[[BaseModel], bool]:
    """
    That decorator is used to get the value of the field from a BaseModel that is provided in the query.
    We do that because we need to get the value of the field in the internal specifications, not just the name
    of it. For example, User(id=1) will use field='id' to get 1 as the result.

    :param func: filtering option function that is going to be decorated
    :param field: field name that is used in getattr(model, field)
    :param value: value of the field
    :return: function to be called with a model to find an attribute and call the comparison function
    """

    @wraps(func)
    def find_attribute_wrapper(model: BaseModel) -> bool:
        foreign_fields = field.split(FILTERING_OPTIONS_SEPARATOR)

        if len(foreign_fields) == 1:
            return func(getattr(model, foreign_fields[0]), value)

        model_val = model
        for foreign_field in foreign_fields:
            if isinstance(model_val, Containers):
                model_val = list(getattr(obj, foreign_field) for obj in model_val)
            elif isinstance(model_val, Dict):
                model_val = list(getattr(obj, foreign_field) for obj in model_val.values())
            else:
                model_val = getattr(model_val, foreign_field)

        if isinstance(model_val, Containers):
            return any(func(member, value) for member in model_val)

        return func(model_val, value)

    find_attribute_wrapper: func
    return find_attribute_wrapper


def eq(field: str, value: Any):
    return find_attribute(func=operator.eq, field=field, value=value)


def gt(field: str, value: Number):
    return find_attribute(func=operator.gt, field=field, value=value)


def gte(field: str, value: Number):
    return find_attribute(func=operator.ge, field=field, value=value)


def lt(field: str, value: Number):
    return find_attribute(func=operator.lt, field=field, value=value)


def lte(field: str, value: Number):
    return find_attribute(func=operator.le, field=field, value=value)


def not_(field: str, value: Union[Literal[True], Literal[False], Literal[None]]):
    return find_attribute(func=operator.not_, field=field, value=value)


def is_(field: str, value: Union[Literal[True], Literal[False], Literal[None]]):
    return find_attribute(func=operator.is_, field=field, value=value)


def regex(field: str, value: str):
    return find_attribute(
        func=lambda model_val, val: re.compile(value).match(model_val),
        field=field,
        value=value,
    )


def like(field: str, value: str):
    return regex(field, f'^{value.replace("%", ".*?")}$')


__all__ = [
    'find_attribute',
    'eq',
    'gt',
    'gte',
    'lt',
    'lte',
    'not_',
    'is_',
    'regex',
    'like',
]



from typing import List, Iterable, Union, Optional, Collection

from assimilator.core.database import specification, SpecificationList, BaseModel
from assimilator.internal.database.specifications.filter_specifications import InternalFilter

QueryT = Union[str, List[BaseModel]]
internal_filter = InternalFilter


@specification
def internal_order(*clauses: str, query: QueryT, **_) -> Iterable[BaseModel]:
    if isinstance(query, str):
        return query

    query = list(query)
    for field in clauses:
        query.sort(
            key=lambda item: getattr(item, field.strip("-")),
            reverse=field.startswith("-"),
        )

    return query


@specification
def internal_paginate(
    *,
    query: QueryT,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    **_,
) -> Iterable[BaseModel]:
    if isinstance(query, str):
        return query

    return list(query)[offset:limit]


@specification
def internal_join(*targets: Collection, query: QueryT, **join_args: dict) -> QueryT:
    return query


@specification
def internal_only(*only_fields: Iterable[str], query: QueryT, **_) -> Iterable[BaseModel]:
    if isinstance(query, str):
        return query

    return (model.copy(include=set(only_fields)) for model in query)


class InternalSpecificationList(SpecificationList):
    filter = internal_filter
    order = internal_order
    paginate = internal_paginate
    join = internal_join
    only = internal_only


__all__ = [
    'internal_filter',
    'InternalFilter',
    'internal_order',
    'internal_paginate',
    'internal_join',
    'internal_only',
    'InternalSpecificationList',
]



from assimilator.core.database.specifications.filtering_options import\
    FilteringOptions, FILTERING_OPTIONS_SEPARATOR


class MongoFilteringOptions(FilteringOptions):
    def _convert_field(self, field: str) -> str:
        return field.replace(FILTERING_OPTIONS_SEPARATOR, ".")

    def _eq(self, field: str, value):
        return {self._convert_field(field): {"$eq": value}}

    def _gt(self, field: str, value):
        return {self._convert_field(field): {"$gt": value}}

    def _gte(self, field: str, value):
        return {self._convert_field(field): {"$gte": value}}

    def _lt(self, field: str, value):
        return {self._convert_field(field): {"$lt": value}}

    def _lte(self, field: str, value):
        return {self._convert_field(field): {"$lte": value}}

    def _not(self, field: str, value):
        return {self._convert_field(field): {"$ne": value}}

    def _is(self, field: str, value):
        return self._eq(self._convert_field(field), value)

    def _like(self, field: str, value):
        return self._regex(field, f'^{value.replace("%", ".*?")}$')

    def _regex(self, field: str, value):
        return {self._convert_field(field): {"$regex": value}}


__all__ = [
    'MongoFilteringOptions',
]



from assimilator.internal.database.specifications.specifications import *
from assimilator.internal.database.specifications.filtering_options import *
from assimilator.internal.database.specifications.internal_operator import *
from assimilator.internal.database.specifications.filter_specifications import *



from redis.client import Pipeline

from assimilator.core.database.unit_of_work import UnitOfWork
from assimilator.redis_.database.repository import RedisRepository


class RedisUnitOfWork(UnitOfWork):
    repository: RedisRepository

    def begin(self):
        self.repository.transaction = self.repository.session.pipeline()

    def rollback(self):
        self.repository.transaction.discard()

    def commit(self):
        self.repository.transaction.execute()

    def close(self):
        self.repository.transaction.reset()
        self.repository.transaction = self.repository.session


__all__ = [
    'RedisUnitOfWork',
]



from typing import Union, Optional, Collection, Type, TypeVar, final

from pymongo import MongoClient

from assimilator.mongo.database.models import MongoModel
from assimilator.core.patterns import LazyCommand, ErrorWrapper
from assimilator.mongo.database.error_wrapper import MongoErrorWrapper
from assimilator.core.database import Repository, SpecificationType, SpecificationList, NotFoundError
from assimilator.mongo.database.specifications.specifications import MongoSpecificationList
from assimilator.internal.database.models_utils import dict_to_models

ModelT = TypeVar("ModelT", bound=MongoModel)


class MongoRepository(Repository):
    session: MongoClient
    model: Type[MongoModel]

    def __init__(
        self,
        session: MongoClient,
        model: Type[MongoModel],
        database: str,
        specifications: Type[SpecificationList] = MongoSpecificationList,
        initial_query: Optional[dict] = None,
        error_wrapper: Optional[ErrorWrapper] = None,
    ):
        super(MongoRepository, self).__init__(
            session=session,
            model=model,
            initial_query=initial_query or {},
            specifications=specifications,
            error_wrapper=error_wrapper or MongoErrorWrapper(),
        )
        self.database = database

    def get_initial_query(self, override_query: Optional[dict] = None) -> dict:
        return dict(super(MongoRepository, self).get_initial_query(override_query))

    @final
    @property
    def _collection(self):
        config = getattr(self.model, 'AssimilatorConfig', None)
        if config is not None:
            collection = self.model.AssimilatorConfig.collection
        else:
            collection = getattr(self.model, 'collection', self.model.__class__.__name__.lower())

        return self.session[self.database][collection]

    def get(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: dict = None,
    ):
        query = self._apply_specifications(
            query=initial_query,
            specifications=specifications,
        )

        data = self._collection.find_one(**query)

        if data is None:
            raise NotFoundError()

        return self.model(**data)

    def filter(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: dict = None
    ) -> Union[Collection[ModelT], LazyCommand[Collection[ModelT]]]:
        query = self._apply_specifications(
            query=initial_query,
            specifications=specifications,
        )
        return [self.model(**data) for data in self._collection.find(**query)]

    def save(self, obj: Optional[ModelT] = None, **obj_data) -> ModelT:
        if obj is None:
            obj = self.model(**dict_to_models(data=obj_data, model=self.model))

        self._collection.insert_one(obj.dict())
        return obj

    def __unpack_query(self, query: dict) -> dict:
        return {
            **query.get("filter", {}),
            **query.get("sort", {}),
            **query.get("skip", {}),
            **query.get("limit", {}),
            **query.get("projection", {}),
        }

    def delete(self, obj: Optional[ModelT] = None, *specifications: SpecificationType) -> None:
        obj, specifications = self._check_obj_is_specification(obj, specifications)

        if specifications:
            query: dict = self._apply_specifications(
                query=self.get_initial_query(),
                specifications=specifications,
            )

            self._collection.delete_many(self.__unpack_query(query))
        elif obj is not None:
            self._collection.delete_one(obj.dict())

    def update(
        self,
        obj: Optional[ModelT] = None,
        *specifications: SpecificationType,
        **update_values,
    ) -> None:
        obj, specifications = self._check_obj_is_specification(obj, specifications)

        if specifications:
            query = self._apply_specifications(
                query=self.get_initial_query(),
                specifications=specifications,
            )

            self._collection.update_many(
                filter=self.__unpack_query(query),
                update={'$set': update_values},
            )
        elif obj is not None:
            self._collection.update_one(
                {"id": obj.id},
                update={'$set': obj.dict()},
                upsert=getattr('obj', 'upsert', False),
            )

    def is_modified(self, obj: ModelT) -> bool:
        return self.get(self.specs.filter(id=obj.id)) == obj

    def refresh(self, obj: ModelT) -> None:
        fresh_obj = self.get(self.specs.filter(id=obj.id))
        obj.__dict__.update(fresh_obj.__dict__)

    def count(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: Optional[dict] = None,
    ) -> Union[LazyCommand[int], int]:
        return self._collection.count_documents(
            filter=self._apply_specifications(
                query=initial_query,
                specifications=specifications,
            ),
        )


__all__ = ['MongoRepository']



from functools import wraps
from abc import ABC, abstractmethod
from typing import (
    TypeVar, Callable, Generic, final,
    Union, Optional, Iterable, Type,
    Collection, Tuple, Any, Dict,
)

from assimilator.core.patterns import ErrorWrapper
from assimilator.core.patterns.lazy_command import LazyCommand
from assimilator.core.database.specifications import SpecificationType, SpecificationList


def make_lazy(func: Callable):

    @wraps(func)
    def make_lazy_wrapper(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: QueryT = None,
    ):
        if lazy:
            return LazyCommand(func, self, *specifications, lazy=False, initial_query=initial_query)
        return func(self, *specifications, lazy=lazy, initial_query=initial_query)

    return make_lazy_wrapper


QueryT = TypeVar("QueryT")
ModelT = TypeVar("ModelT")
SessionT = TypeVar("SessionT")
SpecsT = TypeVar("SpecsT", bound=Type[SpecificationList])


class Repository(Generic[SessionT, ModelT, QueryT], ABC):
    def __init__(
        self,
        session: SessionT,
        model: Type[ModelT],
        specifications: SpecsT,
        initial_query: Optional[QueryT] = None,
        error_wrapper: Optional[ErrorWrapper] = None,
    ):
        self.session = session
        self.model = model
        self.__initial_query: QueryT = initial_query
        self.specifications: SpecsT = specifications

        self.error_wrapper = error_wrapper or ErrorWrapper()
        self.get = LazyCommand.decorate(self.error_wrapper.decorate(self.get))
        self.filter: Repository.filter = LazyCommand.decorate(self.error_wrapper.decorate(self.filter))
        self.save: Repository.save = self.error_wrapper.decorate(self.save)
        self.delete: Repository.delete = self.error_wrapper.decorate(self.delete)
        self.update: Repository.update = self.error_wrapper.decorate(self.update)
        self.is_modified: Repository.is_modified = self.error_wrapper.decorate(self.is_modified)
        self.refresh: Repository.refresh = self.error_wrapper.decorate(self.refresh)
        self.count: Repository.count = LazyCommand.decorate(self.error_wrapper.decorate(self.count))

    @final
    def _check_obj_is_specification(
        self,
        obj: ModelT,
        specifications: Iterable[SpecificationType]
    ) -> Tuple[Optional[ModelT], Iterable[SpecificationType]]:
        """
        This function is called for parts of the code that use both obj and *specifications.
        We check that if the obj is a model
        """

        if not isinstance(obj, self.model) and (obj is not None):
            return None, (obj, *specifications)     # obj is specification

        return obj, specifications

    @property
    def specs(self) -> SpecsT:
        """ That property is used to shorten the full name of the self.specifications. """
        return self.specifications

    def get_initial_query(self, override_query: Optional[QueryT] = None) -> QueryT:
        if override_query is not None:
            return override_query
        elif self.__initial_query is not None:
            return self.__initial_query
        else:
            raise NotImplementedError("You must either pass the initial query or define get_initial_query()")

    def _get_specifications_context(self) -> Dict[str, Any]:
        return {"model": self.model}

    @final
    def _apply_specifications(
        self, query: Union[QueryT, None],
        specifications: Iterable[SpecificationType],
    ) -> QueryT:
        query = self.get_initial_query(query)

        for specification in specifications:
            query = specification(query=query, **self._get_specifications_context())

        return query

    @abstractmethod
    def get(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: QueryT = None,
    ) -> Union[ModelT, LazyCommand[ModelT]]:
        raise NotImplementedError("get() is not implemented()")

    @abstractmethod
    def filter(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: QueryT = None,
    ) -> Union[Collection[ModelT], LazyCommand[Collection[ModelT]]]:
        raise NotImplementedError("filter() is not implemented()")

    @abstractmethod
    def save(self, obj: Optional[ModelT] = None, **obj_data) -> ModelT:
        raise NotImplementedError("save() is not implemented in the repository")

    @abstractmethod
    def delete(self, obj: Optional[ModelT] = None, *specifications: SpecificationType) -> None:
        raise NotImplementedError("delete() is not implemented in the repository")

    @abstractmethod
    def update(self, obj: Optional[ModelT] = None, *specifications: SpecificationType, **update_values) -> None:
        raise NotImplementedError("update() is not implemented in the repository")

    @abstractmethod
    def is_modified(self, obj: ModelT) -> bool:
        raise NotImplementedError("is_modified() is not implemented in the repository")

    @abstractmethod
    def refresh(self, obj: ModelT) -> None:
        raise NotImplementedError("refresh() is not implemented in the repository")

    @abstractmethod
    def count(
        self,
        *specifications: SpecificationType,
        lazy: bool = False,
        initial_query: QueryT = None,
    ) -> Union[LazyCommand[int], int]:
        raise NotImplementedError("count() is not implemented in the repository")

    def __str__(self):
        return f"{self.__class__.__name__}({self.model})"

    def __repr__(self):
        return str(self)


__all__ = [
    'LazyCommand',
    'Repository',
    'make_lazy',
]



from typing import TypeVar, Iterable, Union

from assimilator.core.database import UnitOfWork
from assimilator.core.services.base import Service
from assimilator.core.patterns import LazyCommand

ModelT = TypeVar("ModelT")


class CRUDService(Service):
    def __init__(self, uow: UnitOfWork):
        self.uow = uow
        self._specs = self.uow.repository.specs

    def create(self, obj_data: Union[dict, ModelT]) -> ModelT:
        with self.uow:
            if isinstance(obj_data, dict):
                obj = self.uow.repository.save(**obj_data)
            else:
                obj = self.uow.repository.save(obj_data)

            self.uow.commit()

        self.uow.repository.refresh(obj)
        return obj

    def update(self, update_data: dict, *filters, **kwargs_filters) -> ModelT:
        with self.uow:
            obj = self.get(*filters, **kwargs_filters)

            for key, value in update_data.items():
                setattr(obj, key, value)

            self.uow.repository.update(obj)
            self.uow.commit()

        self.uow.repository.refresh(obj)
        return obj

    def list(
        self, *filters, lazy: bool = False, **kwargs_filters,
    ) -> Union[Iterable[ModelT], LazyCommand[Iterable[ModelT]]]:
        return self.uow.repository.filter(
            self._specs.filter(*filters, **kwargs_filters),
            lazy=lazy,
        )

    def get(self, *filters, lazy: bool = False, **kwargs_filters):
        return self.uow.repository.get(
            self._specs.filter(*filters, **kwargs_filters),
            lazy=lazy,
        )

    def delete(self, *filters, **kwargs_filters):
        with self.uow:
            obj = self.get(*filters, **kwargs_filters)
            self.uow.repository.delete(obj)
            self.uow.commit()

    def __str__(self):
        return f"CRUD({self.uow.repository.model})"


__all__ = [
    'CRUDService',
]



from pymongo.client_session import ClientSession

from assimilator.core.database import UnitOfWork
from assimilator.mongo.database.repository import MongoRepository


class MongoUnitOfWork(UnitOfWork):
    repository: MongoRepository
    transaction: ClientSession

    def begin(self):
        self.transaction = self.repository.session.start_session()
        self.transaction.start_transaction()

    def rollback(self):
        self.transaction.abort_transaction()

    def commit(self):
        self.transaction.commit_transaction()

    def close(self):
        pass


__all__ = [
    'MongoUnitOfWork',
]



from sqlalchemy import column
from sqlalchemy.sql.elements import ColumnClause

from assimilator.core.database.specifications.filtering_options import \
    FilteringOptions, FILTERING_OPTIONS_SEPARATOR


class AlchemyFilteringOptions(FilteringOptions):
    @staticmethod
    def _convert_field(field: str) -> ColumnClause:
        field_parts = field.split(FILTERING_OPTIONS_SEPARATOR)

        if len(field_parts) > 2:
            field = ".".join(field_parts[-2:])
        else:
            field = ".".join(field_parts)

        return column(field, is_literal=True)

    @staticmethod
    def _eq(field, value):
        return AlchemyFilteringOptions._convert_field(field) == value

    @staticmethod
    def _gt(field, value):
        return AlchemyFilteringOptions._convert_field(field) > value

    @staticmethod
    def _gte(field, value):
        return AlchemyFilteringOptions._convert_field(field) >= value

    @staticmethod
    def _lt(field, value):
        return AlchemyFilteringOptions._convert_field(field) < value

    @staticmethod
    def _lte(field, value):
        return AlchemyFilteringOptions._convert_field(field) <= value

    @staticmethod
    def _not(field, value):
        return AlchemyFilteringOptions._convert_field(field) != value

    @staticmethod
    def _is(field, value):
        return AlchemyFilteringOptions._convert_field(field).is_(value)

    @staticmethod
    def _like(field, value):
        return AlchemyFilteringOptions._convert_field(field).like(value)

    @staticmethod
    def _regex(field, value):
        return AlchemyFilteringOptions._convert_field(field).regexp_match(value)


__all__ = [
    'AlchemyFilteringOptions',
]



from abc import abstractmethod, abstractstaticmethod
from typing import Dict, Protocol, Any, Callable


class FilterOptionProtocol(Protocol):
    def __call__(self, field: str, value: Any) -> Callable[[], Any]:
        ...


FILTERING_OPTIONS_SEPARATOR = "__"


class FilteringOptions:
    """ Looks for the filtering option """

    def __init__(self):
        self.filter_options: Dict[str, FilterOptionProtocol] = {
            "eq": self._eq,
            "gt": self._gt,
            "gte": self._gte,
            "lt": self._lt,
            "lte": self._lte,
            "not": self._not,
            "is": self._is,
            "like": self._like,
            "regex": self._regex,
        }

    def get_default_filter(self) -> FilterOptionProtocol:
        return self._eq

    def parse_field(self, raw_field: str, value: Any) -> Callable:
        fields = raw_field.split(FILTERING_OPTIONS_SEPARATOR)
        last_field = fields[-1]

        if len(fields) == 1:
            filter_func = self.filter_options.get(last_field, self.get_default_filter())
            return filter_func(last_field, value)

        # Foreign Key

        if self.filter_options.get(fields[-1]) is None:
            foreign_field = raw_field
            filter_func = self.get_default_filter()
        else:
            foreign_field = FILTERING_OPTIONS_SEPARATOR.join(fields[:-1])
            filter_func = self.filter_options.get(fields[-1])

        return filter_func(foreign_field, value)

    @abstractstaticmethod
    def _eq(field: str, value):
        raise NotImplementedError("_eq() is not implemented in the filtering options")

    @abstractstaticmethod
    def _gt(field: str, value):
        raise NotImplementedError("_gt() is not implemented in the filtering options")

    @abstractstaticmethod
    def _gte(field: str, value):
        raise NotImplementedError("_gte() is not implemented in the filtering options")

    @abstractstaticmethod
    def _lt(field: str, value):
        raise NotImplementedError("_lt() is not implemented in the filtering options")

    @abstractstaticmethod
    def _lte(field: str, value):
        raise NotImplementedError("_lte() is not implemented in the filtering options")

    @abstractstaticmethod
    def _not(field: str, value):
        raise NotImplementedError("_not() is not implemented in the filtering options")

    @abstractstaticmethod
    def _is(field: str, value):
        raise NotImplementedError("_is() is not implemented in the filtering options")

    @abstractstaticmethod
    def _like(field: str, value):
        raise NotImplementedError("_like() is not implemented in the filtering options")

    @abstractstaticmethod
    def _regex(field: str, value):
        raise NotImplementedError("_regex() is not implemented in the filtering options")


__all__ = [
    'FilteringOptions',
    'FilterOptionProtocol',
    'FILTERING_OPTIONS_SEPARATOR',
]



from functools import wraps
from abc import ABC, abstractmethod
from typing import Callable, Union, TypeVar, Type, final, Any

from assimilator.core.database.specifications.filtering_options import FilteringOptions
from assimilator.core.database.specifications.types import (
    OrderSpecificationProtocol,
    PaginateSpecificationProtocol,
    OnlySpecificationProtocol,
    JoinSpecificationProtocol,
)

QueryT = TypeVar("QueryT")


class Specification(ABC):
    @abstractmethod
    def apply(self, query: QueryT, **context: Any) -> QueryT:
        raise NotImplementedError("Specification must specify apply()")

    @final
    def __call__(self, query: QueryT, **context: Any) -> QueryT:
        return self.apply(query, **context)


class FilterSpecification(Specification, ABC):
    filtering_options_cls: Type[FilteringOptions]

    def __init__(self, *filters, **named_filters):
        self.filters = list(filters)
        self.filtering_options = self.filtering_options_cls()

        for field, value in named_filters.items():
            self.filters.append(
                self.filtering_options.parse_field(raw_field=field, value=value)
            )

    def __or__(self, other: 'FilterSpecification') -> 'FilterSpecification':
        raise NotImplementedError("or() is not implemented for FilterSpecification")

    def __and__(self, other: 'FilterSpecification') -> 'FilterSpecification':
        raise NotImplementedError("and() is not implemented for FilterSpecification")

    def __invert__(self):
        raise NotImplementedError("invert() is not implemented for FilterSpecification")

    def __str__(self):
        return f'filter_spec({self.filters})'


def specification(func: Callable) -> Callable:
    def create_specification(*args, **kwargs):
        @wraps(func)
        def created_specification(query: QueryT, **context) -> QueryT:
            return func(*args, **kwargs, query=query, **context)

        created_specification: func
        return created_specification

    return create_specification


SpecificationType = Union[Type[Specification], Callable]


class SpecificationList:
    filter: Type[FilterSpecification]
    order: OrderSpecificationProtocol
    paginate: PaginateSpecificationProtocol
    join: JoinSpecificationProtocol
    only: OnlySpecificationProtocol


__all__ = [
    'SpecificationList',
    'Specification',
    'specification',
    'SpecificationType',
    'FilterSpecification',
]



from abc import abstractmethod
from typing import Iterator, Callable, List, Optional, final

from assimilator.core.events.events import Event
from assimilator.core.patterns.context_managers import StartCloseContextMixin


class EventConsumer(StartCloseContextMixin):
    def __init__(self, callbacks: Optional[List[Callable]] = None):
        if callbacks is None:
            callbacks = []

        self._callbacks: List[Callable] = callbacks

    @final
    def register(self, callback: Callable):
        self._callbacks.append(callback)

    @abstractmethod
    def consume(self) -> Iterator[Event]:
        raise NotImplementedError("consume() is not implemented")


class EventProducer(StartCloseContextMixin):
    @abstractmethod
    def produce(self, event: Event) -> None:
        raise NotImplementedError("produce() is not implemented")


class EventBus:
    def __init__(self, consumer: EventConsumer, producer: EventProducer):
        self.consumer = consumer
        self.producer = producer

    def produce(self, event: Event) -> None:
        self.producer.produce(event)

    def consume(self) -> Iterator[Event]:
        return self.consumer.consume()


__all__ = [
    'EventConsumer',
    'EventProducer',
    'EventBus',
]



import json
from uuid import uuid4
from typing import Type, TypeVar, ClassVar

from pydantic import ValidationError, BaseModel as PydanticBaseModel, Extra

from assimilator.core.exceptions import ParsingError


T = TypeVar("T", bound='BaseModel')


class BaseModel(PydanticBaseModel):
    id: str

    class AssimilatorConfig(PydanticBaseModel, extra=Extra.allow):
        autogenerate_id: ClassVar[bool] = True

    class Config:
        arbitrary_types_allowed = True

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        if not issubclass(cls.AssimilatorConfig, BaseModel.AssimilatorConfig):
            class InheritedConfig(cls.AssimilatorConfig, BaseModel.AssimilatorConfig):
                ...

            cls.AssimilatorConfig = InheritedConfig

        return cls

    def generate_id(self, **kwargs) -> str:
        return str(uuid4())

    def __init__(self, **kwargs):
        if self.AssimilatorConfig.autogenerate_id and kwargs.get('id') is None:
            kwargs['id'] = self.generate_id(**kwargs)

        super(BaseModel, self).__init__(**kwargs)

    @classmethod
    def loads(cls: Type['T'], data: str) -> 'T':
        try:
            return cls(**json.loads(data))
        except (ValidationError, TypeError) as exc:
            raise ParsingError(exc)


__all__ = [
    'BaseModel',
]



from assimilator.core.database.specifications.specifications import *
from assimilator.core.database.specifications.filtering_options import *
from assimilator.core.database.specifications.types import *



from sqlalchemy.exc import NoResultFound, IntegrityError, SQLAlchemyError, MultipleResultsFound

from assimilator.core.database.exceptions import (
    DataLayerError,
    NotFoundError,
    InvalidQueryError,
    MultipleResultsError,
)
from assimilator.core.patterns.error_wrapper import ErrorWrapper


class AlchemyErrorWrapper(ErrorWrapper):
    def __init__(self):
        super(AlchemyErrorWrapper, self).__init__(error_mappings={
            NoResultFound: NotFoundError,
            IntegrityError: InvalidQueryError,
            SQLAlchemyError: DataLayerError,
            MultipleResultsFound: MultipleResultsError,
        }, default_error=DataLayerError)


__all__ = ['AlchemyErrorWrapper']
